@@ -9,12 +9,13 @@ use JQ::Lite::Filters;
 use JQ::Lite::Parser;
 use JQ::Lite::Util ();
 
-our $VERSION = '1.04';
+our $VERSION = '1.05';
 
 sub new {
     my ($class, %opts) = @_;
     my $self = {
         raw => $opts{raw} || 0,
+        _vars => {},
     };
     return bless $self, $class;
 }
@@ -56,7 +57,7 @@ JQ::Lite - A lightweight jq-like JSON query engine in Perl
 
 =head1 VERSION
 
-Version 1.04
+Version 1.05
 
 =head1 SYNOPSIS
 
@@ -93,7 +94,7 @@ jq-like syntax — entirely within Perl, with no external binaries or XS modules
 
 =item * Pipe-style query chaining using | operator
 
-=item * Built-in functions: length, keys, keys_unsorted, values, first, last, reverse, sort, sort_desc, sort_by, min_by, max_by, unique, unique_by, has, contains, any, all, not, group_by, group_count, join, split, explode, implode, count, empty, type, nth, del, delpaths, compact, upper, lower, titlecase, abs, ceil, floor, trim, ltrimstr, rtrimstr, substr, slice, startswith, endswith, add, sum, sum_by, avg_by, median_by, product, min, max, avg, median, mode, percentile, variance, stddev, drop, tail, chunks, range, enumerate, transpose, flatten_all, flatten_depth, clamp, tostring, tojson, fromjson, to_number, pick, merge_objects, to_entries, from_entries, with_entries, map_values, walk, paths, leaf_paths, index, rindex, indices, arrays, objects, scalars
+=item * Built-in functions: length, keys, keys_unsorted, values, first, last, reverse, sort, sort_desc, sort_by, min_by, max_by, unique, unique_by, has, contains, any, all, not, group_by, group_count, join, split, explode, implode, count, empty, type, nth, del, delpaths, compact, upper, lower, titlecase, abs, ceil, floor, trim, ltrimstr, rtrimstr, substr, slice, startswith, endswith, add, sum, sum_by, avg_by, median_by, product, min, max, avg, median, mode, percentile, variance, stddev, drop, tail, chunks, range, enumerate, transpose, flatten_all, flatten_depth, clamp, tostring, tojson, fromjson, to_number, pick, merge_objects, to_entries, from_entries, with_entries, map_values, walk, paths, leaf_paths, getpath, setpath, index, rindex, indices, arrays, objects, scalars
 
 =item * Supports map(...), map_values(...), walk(...), limit(n), drop(n), tail(n), chunks(n), range(...), and enumerate() style transformations
 
@@ -143,6 +144,8 @@ Returns a list of matched results. Each result is a Perl scalar
 
 =item * group_by(.field) (group array items by key)
 
+=item * reduce expr as $var (init; update) (accumulate values with lexical bindings)
+
 =item * group_count(.field) (tally items by key)
 
 =item * sum_by(.field) (sum numeric values projected from each array item)
@@ -288,6 +291,20 @@ Examples:
   .profile | getpath(["emails", 1])     # => "alice.work\@example.com"
   .profile | getpath(paths())
 
+=item * setpath(path; value)
+
+Sets or creates a value at the supplied path, following jq's C<setpath/2>
+semantics. The first argument may be a literal JSON array or any filter that
+emits path arrays. The second argument can be a literal (including JSON
+objects/arrays) or another filter evaluated against the current input. Nested
+hashes/arrays are automatically created as needed, and the original input is
+never mutated.
+
+Examples:
+
+  .settings | setpath(["flags", "beta"]; true)
+  .user     | setpath(["profile", "full_name"]; .name)
+
 =item * pick(key1, key2, ...)
 
 Builds a new object containing only the supplied keys. When applied to arrays
